/* ============================================================
 *
 * This file is a part of the rsbcv project
 *
 * Copyright (C) 2012, 2013, 2014 by Johannes Wienke <jwienke at techfak dot uni-bielefeld dot de>
 *
 * This file may be licensed under the terms of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by:
 *   CoR-Lab, Research Institute for Cognition and Robotics
 *     Bielefeld University
 *
 * ============================================================ */

#include <iostream>

#include <stdlib.h>
#include <math.h>

#include <boost/program_options.hpp>

// mark-start::queue-header
#include <rsc/threading/SynchronizedQueue.h>
// mark-end::queue-header

// mark-start::rsb-headers
#include <rsb/Factory.h>
#include <rsb/Listener.h>
// mark-end::rsb-headers
// mark-start::queuepushhandler
#include <rsb/QueuePushHandler.h>
// mark-end::queuepushhandler
// mark-start::repository-header
#include <rsb/converter/Repository.h>
// mark-end::repository-header
// mark-start::informer-header
#include <rsb/Informer.h>
// mark-end::informer-header

// mark-start::image-converter-header
#include <rst/converters/opencv/IplImageConverter.h>
// mark-end::image-converter-header

// mark-start::rst-headers
#include <rsb/converter/ProtocolBufferConverter.h>
#include <rst/geometry/PointCloud2DInt.pb.h>
// mark-end::rst-headers

// mark-start::opencv-headers
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
// mark-end::opencv-headers

// mark-start::headers-cornerharris
#include "opencv2/imgproc/imgproc.hpp"
// mark-end::headers-cornerharris

using namespace std;
namespace po = boost::program_options;

// mark-start::commandline-arguments
string inScope = "/video";
string outScope = "/results";

void handleCommandline(int argc, char *argv[]) {

    po::options_description options("Allowed options");
    options.add_options()("help,h", "Display a help message.")("inscope,i",
            po::value < string > (&inScope),
            "Scope for receiving input images.")("outscope,o",
            po::value < string > (&outScope), "Scope for sending the results.");

    // allow to give the value as a positional argument
    po::positional_options_description p;
    p.add("value", 1);

    po::variables_map vm;
    po::store(
            po::command_line_parser(argc, argv).options(options).positional(p).run(),
            vm);

    // first, process the help option
    if (vm.count("help")) {
        cout << options << "\n";
        exit(1);
    }

    // afterwards, let program options handle argument errors
    po::notify(vm);

}
// mark-end::commandline-arguments

int main(int argc, char *argv[]) {

    handleCommandline(argc, argv);

    // mark-start::converter-registration
    rsb::converter::Converter<string>::Ptr imageConverter(
            new rst::converters::opencv::IplImageConverter());
    rsb::converter::converterRepository<string>()->registerConverter(imageConverter);
    // mark-end::converter-registration

    // mark-start::rst-converter-registration
    rsb::converter::Converter<string>::Ptr pointCloudConverter(
            new rsb::converter::ProtocolBufferConverter<rst::geometry::PointCloud2DInt> ());
    rsb::converter::converterRepository<string>()->registerConverter(pointCloudConverter);
    // mark-end::rst-converter-registration

    // mark-start::factory
    rsb::Factory& factory = rsb::getFactory();
    // mark-end::factory

    // mark-start::listener
    rsb::ListenerPtr imageListener = factory.createListener(
            rsb::Scope(inScope));
    // mark-end::listener

    // mark-start::create-queue
    boost::shared_ptr<rsc::threading::SynchronizedQueue<boost::shared_ptr<IplImage> > > imageQueue(
                    new rsc::threading::SynchronizedQueue<boost::shared_ptr<IplImage> >(1));
    // mark-end::create-queue

    // mark-start::handler
    imageListener->addHandler(
            rsb::HandlerPtr(new rsb::QueuePushHandler<IplImage>(imageQueue)));
    // mark-end::handler

    // mark-start::informer-instance
    rsb::Informer<rst::geometry::PointCloud2DInt>::Ptr informer =
            factory.createInformer<rst::geometry::PointCloud2DInt>(rsb::Scope(outScope));
    // mark-end::informer-instance

    // mark-start::working-loop
    while (true) {

        cv::Mat image = cv::Mat(imageQueue->pop().get(), true);

        cv::imshow("input", image);
        cv::waitKey(1);
        // mark-end::working-loop

        // mark-start::cornerharris
        cv::Mat image_gray;
        cvtColor(image, image_gray, CV_BGR2GRAY);
        cv::Mat dst = cv::Mat::zeros(image_gray.size(), CV_32FC1);
        cv::cornerHarris(image_gray, dst, 2, 3, 0.04);
        // mark-end::cornerharris

        // mark-start::display-keypoints
        cv::Mat dst_norm, dst_norm_scaled;
        // normalize and scale the image
        cv::normalize(dst, dst_norm, 0, 255, cv::NORM_MINMAX, CV_32FC1,
                cv::Mat());
        cv::convertScaleAbs(dst_norm, dst_norm_scaled);

        // The target
        cv::Mat imageWithKeyPoints = image;
        // draw a circle around corners
        for (int j = 0; j < dst_norm.rows; j++) {
            for (int i = 0; i < dst_norm.cols; i++) {
                // a somewhat good threshold
                if (dst_norm.at<float>(j, i) > 195) {
                    // Circle the result
                    circle(imageWithKeyPoints, cv::Point(i, j), 5,
                            cv::Scalar(0), 2, 8, 0);
                }
            }
        }

        cv::imshow("keypoints", imageWithKeyPoints);
        cv::waitKey(1);
        // mark-end::display-keypoints

        // mark-start::publish
        boost::shared_ptr<rst::geometry::PointCloud2DInt> cloud(
                new rst::geometry::PointCloud2DInt());
        for (int j = 0; j < dst_norm.rows; j++) {
            for (int i = 0; i < dst_norm.cols; i++) {
                // a somewhat good threshold
                if (dst_norm.at<float>(j, i) > 195) {
                    rst::math::Vec2DInt *point = cloud->add_points();
                    point->set_x(i);
                    point->set_y(j);
                }
            }
        }
        informer->publish(cloud);
        // mark-end::publish

    }

    return EXIT_SUCCESS;

}
